"*timeout:		10",
"*cycle:			10",
"*lockTimeout:		0",
"*passwdTimeout:		30",
"*nice:			10",
"*lock:			False",
"*lockVTs:		True",
"*verbose:		False",
"*timestamp:		False",
"*fade:			True",
"*unfade:		False",
"*fadeSeconds:		3",
"*fadeTicks:		20",
"*splash:		True",
"*splashDuration:	5",
"*visualID:		default",
"*captureStderr: 	True",
"*overlayTextForeground:	#FFFF00",
"*overlayTextBackground:	#000000",
"*overlayStderr:		True",
"*font:			*-medium-r-*-140-*-m-*",
"*sgiSaverExtension:	True",
"*mitSaverExtension:	False",
"*xidleExtension:	True",
"*procInterrupts:	True",
"*demoCommand: xscreensaver-demo",
"*prefsCommand: xscreensaver-demo -prefs",
"*helpURL: http://www.jwz.org/xscreensaver/man.html",
"*loadURL: netscape -remote 'openURL(%s)' || netscape '%s'",
"*installColormap:	True",
"*programs:	qix -root -solid -delay 0 -segments 100			\\n\
		attraction -root -mode balls				\\n\
		attraction -root -mode lines -points 3 -segments 200	\\n\
		attraction -root -mode splines -segments 300		\\n\
		attraction -root -mode lines -radius 300		  \
			-orbit -vmult 0.5				\\n\
		pyro -root						\\n\
		helix -root						\\n\
		pedal -root						\\n\
		rorschach -root -offset 7				\\n\
		hopalong -root						\\n\
		greynetic -root						\\n\
		xroger -root						\\n\
		imsmap -root						\\n\
		slidescreen -root					\\n\
		decayscreen -root					\\n\
		jigsaw -root						\\n\
		blitspin -root -grab					\\n\
		slip -root						\\n\
		distort -root						\\n\
		spotlight -root						\\n\
		hypercube -root						\\n\
		halo -root						\\n\
		maze -root						\\n\
		noseguy -root						\\n\
		flame -root						\\n\
		lmorph -root						\\n\
		deco -root						\\n\
		moire -root						\\n\
		moire2 -root						\\n\
		lightning -root						\\n\
		strange -root						\\n\
		spiral -root						\\n\
		laser -root						\\n\
		grav -root						\\n\
		grav -root -trail -decay				\\n\
		drift -root						\\n\
		ifs -root						\\n\
		julia -root						\\n\
		penrose -root						\\n\
		sierpinski -root					\\n\
		braid -root						\\n\
		galaxy -root						\\n\
		bouboule -root						\\n\
		swirl -root						\\n\
		flag -root						\\n\
		sphere -root						\\n\
		forest -root						\\n\
		lisa -root						\\n\
		lissie -root						\\n\
		goop -root						\\n\
		starfish -root						\\n\
		starfish -root -blob					\\n\
		munch -root						\\n\
		fadeplot -root						\\n\
		coral -root						\\n\
		mountain -root						\\n\
		triangle -root						\\n\
		worm -root						\\n\
		rotor -root						\\n\
		ant -root						\\n\
		demon -root						\\n\
		loop -root						\\n\
		vines -root						\\n\
		kaleidescope -root					\\n\
		xjack -root						\\n\
  -		xlyap -root -randomize					\\n\
		cynosure -root						\\n\
		flow -root						\\n\
		epicycle -root						\\n\
		interference -root					\\n\
		truchet -root -randomize				\\n\
		bsod -root						\\n\
		crystal -root						\\n\
		discrete -root						\\n\
		kumppa -root						\\n\
		rd-bomb -root						\\n\
		rd-bomb -root -speed 1 -size 0.1			\\n\
		sonar -root						\\n\
		t3d -root						\\n\
		penetrate -root						\\n\
		deluxe -root						\\n\
		compass -root						\\n\
		squiral -root						\\n\
		xflame -root						\\n\
		wander -root						\\n\
		wander -root -advance 0 -size 10 -circles True		  \
		  -length 10000 -reset 100000				\\n\
		critical -root						\\n\
		phosphor -root						\\n\
		xmatrix -root						\\n\
									  \
	mono:	rocks -root						\\n\
	color:	rocks -root -fg darksalmon				\\n\
									  \
	mono:	qix -root -linear -count 5 -size 200 -spread 30		  \
			-segments 75 -solid -xor			\\n\
									  \
	color:	attraction -root -mode polygons				\\n\
	color:	attraction -root -mode filled-splines -segments 0	\\n\
	color:	attraction -root -glow -points 10			\\n\
	color:	bubbles -root						\\n\
									  \
  PseudoColor:	qix -root -count 4 -solid -transparent			\\n\
  PseudoColor:	qix -root -count 5 -solid -transparent -linear		  \
			-segments 250 -size 100				\\n\
									  \
	   GL:	gears -root					  	\\n\
	   GL:	superquadrics -root				  	\\n\
	   GL:	morph3d -root					  	\\n\
	   GL:	cage -root					  	\\n\
	   GL:	moebius -root					  	\\n\
	   GL:	stairs -root					  	\\n\
	   GL:	pipes -root					  	\\n\
	   GL:	sproingies -root				  	\\n\
	   GL:	rubik -root					  	\\n\
	   GL:	atlantis -root					  	\\n\
	   GL:	lament -root					  	\\n\
	   GL:	bubble3d -root					  	\\n\
	   GL:	glplanet -root					  	\\n\
	   GL:	pulsar -root					  	\\n\
  -	   GL:  pulsar -root -texture -mipmap -texture_quality  	  \
		       -light -fog					\\n",
" ",
"XScreenSaver.pointerPollTime:		5",
"XScreenSaver.initialDelay:		0",
"XScreenSaver.windowCreationTimeout:	30",
"XScreenSaver.bourneShell:		/bin/sh",
"*Dialog.headingFont:		*-times-bold-r-*-*-*-180-*-*-*-iso8859-1",
"*Dialog.bodyFont:		*-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"*Dialog.labelFont:		*-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"*Dialog.buttonFont:		*-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"*Dialog.foreground:		#000000",
"*Dialog.background:		#BFBFBF",
"*Dialog.Button.foreground:	#000000",
"*Dialog.Button.background:	#D0D0D0",
"*Dialog.text.foreground:	#000000",
"*Dialog.text.background:	#FFFFFF",
"*Dialog.logo.foreground:	#FF0000",
"*Dialog.logo.background:	#FFFFFF",
"*Dialog.topShadowColor:		#E7E7E7",
"*Dialog.bottomShadowColor:	#737373",
"*Dialog.logo.width:		200",
"*Dialog.logo.height:		200",
"*Dialog.internalBorderWidth:	30",
"*Dialog.borderWidth:		1",
"*Dialog.shadowThickness:	4",
"*passwd.heading.label:		XScreenSaver %s",
"*passwd.body.label:		This display is locked.",
"*passwd.user.label:		User:",
"*passwd.passwd.label:		Password:",
"*passwd.passwdFont:		*-courier-medium-r-*-*-*-140-*-*-*-iso8859-1",
"*passwd.thermometer.width:	8",
"*splash.heading.label:		XScreenSaver %s",
"*splash.body.label:		Copyright © 1991-1999 by",
"*splash.body2.label:		Jamie Zawinski <jwz@jwz.org>",
"*splash.demo.label:		Demo",
"*splash.prefs.label:		Prefs",
"*splash.help.label:		Help",
"*fontList:                       *-helvetica-medium-r-*-*-*-120-*-*-*-iso8859-1",
"*demoDialog*label1.fontList:     *-helvetica-medium-r-*-*-*-140-*-*-*-iso8859-1",
"*XmTextField.fontList:             *-courier-medium-r-*-*-*-120-*-*-*-iso8859-1",
"*label0.fontList:                  *-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"XScreenSaver*XmList.fontList:      *-courier-medium-r-*-*-*-120-*-*-*-iso8859-1",
"*XmDialogShell*foreground:		#000000",
"*XmDialogShell*background:		#E5E5E5",
"*XmDialogShell*XmTextField.foreground:	#000000",
"*XmDialogShell*XmTextField.background:	#FFFFFF",
"*XmDialogShell*demoList.foreground:	#000000",
"*XmDialogShell*demoList.background:	#FFFFFF",
"*XmDialogShell.title:		XScreenSaver",
"*versionWarning_popup.title:	XScreenSaver Warning",
"*demoForm_popup.title:		XScreenSaver Demo",
"*preferencesForm_popup.title:	XScreenSaver Preferences",
"*allowShellResize:		True",
"*autoUnmanage:			False",
"*demoDialog.maxWidth:		600",
"*label1.labelString:		XScreenSaver %s",
"*label1.label:			XScreenSaver %s",
"*label2.labelString: Copyright © 1991-1999 by Jamie Zawinski <jwz@jwz.org>",
"*label2.label:	     Copyright © 1991-1999 by Jamie Zawinski <jwz@jwz.org>",
"*demoList.visibleItemCount:	10",
"*demoList.automaticSelection:	True",
"*next.labelString:		Run Next",
"*prev.labelString:		Run Previous",
"*edit.labelString:		Preferences",
"*restart.labelString:		Reinitialize",
"*done.labelString:		Quit",
"*preferencesLabel.labelString:	XScreenSaver Parameters",
"*timeoutLabel.labelString:	Saver Timeout",
"*cycleLabel.labelString:	Cycle Timeout",
"*fadeSecondsLabel.labelString:	Fade Duration",
"*fadeTicksLabel.labelString:	Fade Ticks",
"*lockLabel.labelString:		Lock Timeout",
"*passwdLabel.labelString:	Password Timeout",
"*preferencesForm*XmTextField.columns:	8",
"*verboseToggle.labelString:	Verbose",
"*cmapToggle.labelString:	Install Colormap",
"*fadeToggle.labelString:	Fade Colormap",
"*unfadeToggle.labelString:	Unfade Colormap",
"*lockToggle.labelString:	Require Password",
"*preferencesDone.labelString:	OK",
"*preferencesCancel.labelString:	Cancel",
"XScreenSaver*dragInitiatorProtocolStyle: DRAG_NONE",
"XScreenSaver*dragReceiverProtocolStyle:  DRAG_NONE",
"*demo_dialog.title:		XScreenSaver Demo",
"*preferences_dialog.title:	XScreenSaver Preferences",
"*warning_dialog.title:		XScreenSaver Warning",
"*demo_dialog.geometry:		=640x400",
"*demo_dialog*font:		*-helvetica-bold-r-*-*-*-120-*-*-*-iso8859-1",
"*preferences_dialog*font:	*-helvetica-bold-r-*-*-*-120-*-*-*-iso8859-1",
"*demo_dialog*label1.font:	*-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"*preferences_dialog*label1.font:*-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"XScreenSaver*warning_dialog*label0.font:	\
				*-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"XScreenSaver*warning_dialog*Label.font:	\
				*-helvetica-bold-r-*-*-*-120-*-*-*-iso8859-1",
"XScreenSaver*warning_dialog*Command.font: \
				*-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"XScreenSaver.demo_dialog*List.font:	\
				*-courier-medium-r-*-*-*-120-*-*-*-iso8859-1",
"XScreenSaver.demo_dialog*Text*font:	\
				*-courier-medium-r-*-*-*-120-*-*-*-iso8859-1",
"XScreenSaver.demo_dialog*foreground:			#000000",
"XScreenSaver.demo_dialog*background:			#E5E5E5",
"XScreenSaver.demo_dialog*List.background:		#FFFFFF",
"XScreenSaver.demo_dialog*Scrollbar.background:		#D9D9D9",
"XScreenSaver.demo_dialog*Command.background:		#D9D9D9",
"XScreenSaver.demo_dialog*Text*background:		#FFFFFF",
"XScreenSaver.preferences_dialog*foreground:		#000000",
"XScreenSaver.preferences_dialog*background:		#E5E5E5",
"XScreenSaver.preferences_dialog*Command.background:	#D9D9D9",
"XScreenSaver.preferences_dialog*Toggle.background:	#D9D9D9",
"XScreenSaver.preferences_dialog*Text*background:	#FFFFFF",
"XScreenSaver.warning_dialog*foreground:			#000000",
"XScreenSaver.warning_dialog*background:			#E5E5E5",
"XScreenSaver.warning_dialog*Command.background:		#D9D9D9",
"*preferences_dialog*Dialog.value.translations: #override\\n\
	<Key>Return: beginning-of-line()\\n",
"*demo_dialog*viewport.height:			200",
"*Form.borderWidth:				0",
"*Box.borderWidth:				0",
"*Label.borderWidth:				0",
"*preferences_dialog*Dialog.borderWidth:		0",
"*demo_dialog*next.label:			Run Next",
"*demo_dialog*prev.label:			Run Previous",
"*demo_dialog*edit.label:			Preferences",
"*demo_dialog*restart.label:			Reinitialize",
"*demo_dialog*done.label:			Quit",
"XScreenSaver.demo_dialog*Command.internalWidth:  10",
"XScreenSaver.demo_dialog*Command.internalHeight: 4",
"*preferences_dialog*timeout.label:		Saver Timeout:",
"*preferences_dialog*cycle.label:		Cycle Timeout:",
"*preferences_dialog*fade.label:			Fade Duration:",
"*preferences_dialog*ticks.label:		Fade Ticks:",
"*preferences_dialog*lockTime.label:		Lock Timeout:",
"*preferences_dialog*passwdTime.label:		Password Timeout:",
"XScreenSaver.preferences_dialog*Command.internalWidth:  10",
"XScreenSaver.preferences_dialog*Command.internalHeight: 4",
"*preferences_dialog*label1.label:		XScreenSaver Parameters",
"*preferences_dialog*buttonbox.verbose.label:	Verbose",
"*preferences_dialog*buttonbox.cmap.label:	Install Colormap",
"*preferences_dialog*buttonbox.fade.label:	Fade Colormap",
"*preferences_dialog*buttonbox.unfade.label:	Unfade Colormap",
"*preferences_dialog*buttonbox.lock.label:	Require Password",
"*preferences_dialog*done.label:			Ok",
"*preferences_dialog*cancel.label:		Cancel",
"*warning_dialog*ok.label:			Ok",
"*warning_dialog*horizDistance:			30",
"*warning_dialog*vertDistance:			0",
"*warning_dialog*Label.internalWidth:		1",
"*warning_dialog*Label.internalHeight:		0",
"*warning_dialog*label0.horizDistance:		80",
"*warning_dialog*label0.vertDistance:		20",
"*warning_dialog*Command.horizDistance:		160",
"*warning_dialog*Command.vertDistance:		20",
"*warning_dialog*Command.internalWidth:		20",
"*warning_dialog*Command.internalHeight:		5",
